"""
Copyright 2020 ShipChain, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import logging

import requests
from aws_requests_auth.boto_utils import BotoAWSRequestsAuth
from django.conf import settings
from influxdb_metrics.loader import log_metric, TimingMetric
from rest_framework import status

from .exceptions import AWSIoTError
from .utils import DecimalEncoder

LOG = logging.getLogger('python-common')


class AWSClient:
    @property
    def url(self):
        raise NotImplementedError

    @property
    def session(self):
        raise NotImplementedError

    METHOD_POST = 'post'
    METHOD_PUT = 'put'
    METHOD_GET = 'get'
    METHOD_DELETE = 'delete'

    RESPONSE_200_METHODS = [METHOD_PUT, METHOD_GET, METHOD_DELETE]

    def _call(self, http_method, endpoint, payload=None, params=None):
        metric_name = self._get_generic_endpoint_for_metric(http_method, endpoint)
        calling_url = f'{self.url}/{endpoint}'

        if payload:
            payload = json.dumps(payload, cls=DecimalEncoder)

        try:

            with TimingMetric('python_common_aws.call', tags={'method': metric_name}) as timer:

                if http_method == self.METHOD_POST:
                    response = self.session.post(calling_url, data=payload, params=params)
                    response_json = response.json()

                    if response.status_code != status.HTTP_201_CREATED:
                        self._process_error_object(metric_name, response, response_json)

                elif http_method in self.RESPONSE_200_METHODS:
                    response = getattr(self.session, http_method)(calling_url, data=payload, params=params)
                    response_json = response.json()

                    if response.status_code != status.HTTP_200_OK:
                        self._process_error_object(metric_name, response, response_json)

                else:
                    log_metric('python_common_aws.error', tags={'method': metric_name, 'code': 'InvalidHTTPMethod'})
                    LOG.error('aws_client(%s) error: %s', metric_name, 'Invalid HTTP Method')
                    raise AWSIoTError(f'Invalid HTTP Method {http_method}')

                LOG.info('aws_client(%s) duration: %.3f', metric_name, timer.elapsed)

        except requests.exceptions.ConnectionError:
            log_metric('python_common_aws.error', tags={'method': metric_name, 'code': 'ConnectionError'})
            raise AWSIoTError("Service temporarily unavailable, try again later", status.HTTP_503_SERVICE_UNAVAILABLE,
                              'service_unavailable')

        except Exception as exception:
            log_metric('python_common_aws.error', tags={'method': metric_name, 'code': 'exception'})
            raise AWSIoTError(str(exception))

        return response_json

    def _post(self, endpoint='', payload=None, query_params=None):
        return self._call(self.METHOD_POST, endpoint, payload, params=query_params)

    def _put(self, endpoint='', payload=None, query_params=None):
        return self._call(self.METHOD_PUT, endpoint, payload, params=query_params)

    def _get(self, endpoint='', query_params=None):
        return self._call(self.METHOD_GET, endpoint, params=query_params)

    def _delete(self, endpoint='', query_params=None):
        return self._call(self.METHOD_DELETE, endpoint, params=query_params)

    @staticmethod
    def _process_error_object(endpoint, response, response_json):
        error_code = response.status_code

        if 'error' in response_json:
            message = response_json['error']
            if isinstance(message, dict):
                if 'code' in message:
                    error_code = message['code']
                if 'message' in message:
                    message = message['message']

        elif 'message' in response_json:
            message = response_json['message']

        else:
            message = response_json

        log_metric('python_common_aws.error', tags={'method': endpoint, 'code': error_code})
        LOG.error('aws_client(%s) error: %s', endpoint, message)
        raise AWSIoTError(f'Error in AWS IoT Request: [{error_code}] {message}')

    def _get_generic_endpoint_for_metric(self, http_method, endpoint):
        # This should be overwritten by each usage of this class for added clarity.
        return f'{http_method}::{endpoint}'


class URLShortenerClient(AWSClient):
    url = settings.URL_SHORTENER_URL
    session = requests.session()

    def __init__(self):
        aws_auth = BotoAWSRequestsAuth(
            aws_host=settings.URL_SHORTENER_HOST,
            aws_region='us-east-1',
            aws_service='execute-api'
        )

        self.session.headers = {'content-type': 'application/json'}
        self.session.auth = aws_auth

    def _get_generic_endpoint_for_metric(self, http_method, endpoint):
        return f'urlshortener::{http_method}::{endpoint}'



"""
Copyright 2019 ShipChain, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import re

import requests
from aws_requests_auth.boto_utils import BotoAWSRequestsAuth
from django.conf import settings

from .aws import AWSClient

LOG = logging.getLogger('python-common')


class AWSIoTClient(AWSClient):
    url = f'https://{settings.IOT_AWS_HOST}/{settings.IOT_GATEWAY_STAGE}'
    session = requests.session()

    def __init__(self):
        aws_auth = BotoAWSRequestsAuth(
            aws_host=settings.IOT_AWS_HOST,
            aws_region='us-east-1',
            aws_service='execute-api'
        )

        self.session.headers = {'content-type': 'application/json'}
        self.session.auth = aws_auth

    def _get_generic_endpoint_for_metric(self, http_method, endpoint):
        generic_endpoint = re.sub(r'[0-9A-F]{8}-[0-9A-F]{4}-[4][0-9A-F]{3}-[89AB][0-9A-F]{3}-[0-9A-F]{12}',
                                  '<device_id>', endpoint, flags=re.IGNORECASE)

        return f'iot::{http_method}::{generic_endpoint}'



"""
Copyright 2019 ShipChain, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pytest
from rest_framework import status

# pylint: disable=too-many-branches
# pylint: disable=too-many-arguments


class EntityReferenceClass:
    def __init__(self, resource=None, pk=None, attributes=None, relationships=None):
        self.resource = resource
        self.pk = pk
        self.attributes = attributes
        self.relationships = relationships

    def __str__(self):
        return f'Type: {self.resource}; ID: {self.pk}; ' \
               f'attributes: {self.attributes}; relationships: {self.relationships}'


def _vnd_has_error(response_json, error, pointer=None):
    assert 'errors' in response_json, f'Malformed error response: {response_json}'
    errors = response_json['errors']
    assert isinstance(errors, list), f'Error response not a list: {errors}'
    error_found = False

    for single_error in errors:
        if error in single_error['detail']:
            error_found = True
            if pointer:
                found_pointer = single_error['source']['pointer']
                assert pointer in found_pointer, f'Error `{pointer}` not found in {found_pointer}'

    if not error_found:
        assert False, f'Error `{error}` not found in {errors}'


def _json_has_error(errors, error):
    assert isinstance(errors, dict), f'Error response not a dict: {errors}'
    assert 'detail' in errors, f'Malformed error response: {errors}'

    assert errors['detail'] == error, f'Error {error} not found in {errors["detail"]}'


def response_has_error(response, error, pointer=None, vnd=True):
    if error is None:
        return

    response_json = response.json()

    # application/vnd.api+json
    if vnd:
        _vnd_has_error(response_json, error, pointer)

    # application/json
    else:
        _json_has_error(response_json, error)


def _vnd_assert_attributes(response_data, attributes):
    """
    Scan response data for all attributes
    """
    assert 'attributes' in response_data, f'Attributes missing in {response_data}'
    response_attributes = response_data['attributes']

    for key, value in attributes.items():
        assert key in response_attributes, f'Missing Attribute `{key}` in {response_attributes}'
        assert response_attributes[key] == value, f'Attribute Value incorrect `{value}` in {response_attributes}'


def _vnd_assert_entity_ref_in_list(response_list, entity_ref, skip_attributes_property=False):
    found_include = False

    if entity_ref.attributes is None:
        entity_ref.attributes = dict()

    for response_single in response_list:
        if entity_ref.resource and entity_ref.pk:
            if response_single['type'] == entity_ref.resource and response_single['id'] == entity_ref.pk:
                found_include = True
                if skip_attributes_property:
                    break
                for attr_key, attr_value in entity_ref.attributes.items():
                    assert attr_key in response_single['attributes'], \
                        f'List Attribute key `{attr_key}` missing in {response_single}'
                    assert response_single['attributes'][attr_key] == attr_value, \
                        f'List Attribute Value incorrect `{attr_value}` in {response_single}'
                break
        else:
            single_attribute_failed = False

            for attr_key, attr_value in entity_ref.attributes.items():
                if attr_key not in response_single['attributes']:
                    single_attribute_failed = True
                    break
                elif response_single['attributes'][attr_key] != attr_value:
                    single_attribute_failed = True
                    break

            if not single_attribute_failed:
                found_include = True

    assert found_include, f'{entity_ref} NOT IN  {response_list}'


def _vnd_assert_relationships(response_data, relationships):
    """
    Scan response data for all relationships
    """
    assert 'relationships' in response_data, f'Relationships missing in {response_data}'
    response_relationships = response_data['relationships']

    if not isinstance(relationships, list):
        relationships = [relationships]

    for relationship in relationships:
        for relationship_name, relationship_refs in relationship.items():

            assert relationship_name in response_relationships, \
                f'Relationship `{relationship_name}` not in {response_relationships}'

            if not isinstance(relationship_refs, list):
                relationship_refs = [relationship_refs]

            for relationship_ref in relationship_refs:

                assert isinstance(relationship_ref, EntityReferenceClass), \
                    f'asserted relationship is not an EntityRef {relationship_ref}'

                assert 'data' in response_relationships[relationship_name], \
                    f'Data missing in {relationship_name} relationship : {response_relationships[relationship_name]}'

                if isinstance(response_relationships[relationship_name]['data'], list):
                    _vnd_assert_entity_ref_in_list(response_relationships[relationship_name]['data'], relationship_ref,
                                                   skip_attributes_property=True)

                else:
                    if relationship_ref.resource:
                        assert response_relationships[relationship_name]['data']['type'] == relationship_ref.resource, \
                            f'EntityRef resource type `{relationship_ref.resource}` does not match {response_relationships}'
                    if relationship_ref.pk:
                        assert response_relationships[relationship_name]['data']['id'] == relationship_ref.pk, \
                            f'EntityRef ID `{relationship_ref.pk}` does not match {response_relationships}'


def _vnd_assert_include(response, included):
    """
    Scan a response for all included resources
    """
    assert 'included' in response, f'Included missing in {response}'
    response_included = response['included']

    if not isinstance(included, list):
        included = [included]

    for single_include in included:
        assert isinstance(single_include, EntityReferenceClass), \
            f'asserted includes is not an EntityRef {single_include}'
        _vnd_assert_entity_ref_in_list(response_included, single_include)


def _vnd_assertions(response_data, entity_ref):
    if entity_ref.resource:
        assert response_data['type'] == entity_ref.resource, f'Invalid Resource Type in {response_data}'

    if entity_ref.pk:
        assert response_data['id'] == entity_ref.pk, f'Invalid ID in {response_data}'

    if entity_ref.attributes:
        _vnd_assert_attributes(response_data, entity_ref.attributes)

    if entity_ref.relationships:
        _vnd_assert_relationships(response_data, entity_ref.relationships)


def _plain_assert_attributes_in_response(response, attributes):
    for key, value in attributes.items():
        assert key in response, f'Missing Attribute `{key}` in {response}'
        if isinstance(value, dict):
            _plain_assert_attributes_in_response(response[key], value)
        else:
            assert response[key] == value, f'Attribute Value incorrect `{value}` in {response}'


def _plain_assert_attributes_in_list(response_list, attributes):
    found_include = False

    if attributes is None:
        attributes = dict()

    for response_single in response_list:
        single_attribute_failed = False

        try:
            _plain_assert_attributes_in_response(response_single, attributes)
        except AssertionError:
            single_attribute_failed = True

        if not single_attribute_failed:
            found_include = True

    assert found_include, f'{attributes} NOT IN  {response_list}'


def _test_vnd_json(response, entity_refs=None, included=None, is_list=False, count=None, resource=None, pk=None,
                   attributes=None, relationships=None, check_ordering=False):
    assert 'data' in response, f'response does not contain `data` property: {response}'

    # if (attributes or relationships or resource or pk) and entity_refs:
    assert not ((attributes or relationships or resource or pk) and entity_refs), \
        'Use Only `entity_refs` or explicit `attributes`, `relationships`, `resource`, and `pk` but not both'

    if (attributes or relationships or resource or pk) and not entity_refs:
        entity_refs = EntityReferenceClass(resource=resource,
                                           pk=pk,
                                           attributes=attributes,
                                           relationships=relationships)

    response_data = response['data']

    if is_list:
        assert isinstance(response_data, list), f'Response should be a list'

        # Included resources are outside of the list response
        if included:
            _vnd_assert_include(response, included)

        # Assertion for only included and not entities is valid
        if entity_refs:
            if not isinstance(entity_refs, list):
                entity_refs = [entity_refs]

            if not check_ordering:
                for entity_ref in entity_refs:
                    _vnd_assert_entity_ref_in_list(response_data, entity_ref)
            else:
                assert len(entity_refs) <= len(response_data), \
                    f'Error: more entity refs supplied than available in response data. ' \
                    f'{len(response_data)} found asserted {len(entity_refs)}'
                for iteration, entity_ref in enumerate(entity_refs):
                    _vnd_assertions(response_data[iteration], entity_ref)

        if count is not None:
            assert len(response_data) == count, \
                f'Difference in count of response_data, got {len(response_data)} expected {count}'

    else:
        assert not isinstance(response_data, list), f'Response should not be a list'
        assert not (entity_refs and isinstance(entity_refs, list)), \
            f'entity_refs should not be a list for a non-list response'

        assert (count is None), 'Count is only checked when response is list'
        assert not check_ordering, 'Ordering is only checked when response is list'

        # Included resources are outside of the list response
        if included:
            _vnd_assert_include(response, included)

        # Assertion for only status is valid
        if entity_refs:
            _vnd_assertions(response_data, entity_refs)


def _test_regular_json(response, entity_refs=None, included=None, is_list=False, count=None, attributes=None,
                       relationships=None, check_ordering=False):
    assert not relationships, f'relationships not valid when vnd=False'
    assert not entity_refs, f'entity_refs not valid when vnd=False'
    assert not included, f'included not valid when vnd=False'
    assert attributes, f'attributes must be provided when vnd=False'

    if is_list:
        assert isinstance(response, list), f'Response should be a list'

        if not isinstance(attributes, list):
            attributes = [attributes]

        if not check_ordering:
            for attribute in attributes:
                _plain_assert_attributes_in_list(response, attribute)

        else:
            assert len(attributes) <= len(response), \
                f'Error: more attributes supplied than available in response. ' \
                f'{len(attributes)} found asserted {len(response)}'
            for iteration, attribute in enumerate(attributes):
                _plain_assert_attributes_in_response(response[iteration], attribute)
        if count is not None:
            assert len(response) == count,\
                f'Difference in count of response_data, got {len(response)} expected {count}'
    else:
        assert not isinstance(response, list), f'Response should not be a list'
        assert not (attributes and isinstance(attributes, list)), \
            f'attributes should not be a list for a non-list response'

        _plain_assert_attributes_in_response(response, attributes)


def response_has_data(response, vnd=True, entity_refs=None, included=None, is_list=False, count=None,
                      resource=None, pk=None, attributes=None, relationships=None, check_ordering=False):
    response = response.json()

    # application/vnd.api+json
    if vnd:
        _test_vnd_json(response, entity_refs, included, is_list, count, resource, pk, attributes, relationships,
                       check_ordering)

    # application/json
    else:
        _test_regular_json(response, entity_refs, included, is_list, count, attributes, relationships, check_ordering)


def assert_200(response, vnd=True, entity_refs=None, included=None, is_list=False, count=None, check_ordering=False,
               resource=None, pk=None, attributes=None, relationships=None):
    assert response is not None
    assert response.status_code == status.HTTP_200_OK, f'status_code {response.status_code} != 200'
    response_has_data(response,
                      attributes=attributes,
                      relationships=relationships,
                      included=included,
                      is_list=is_list,
                      vnd=vnd,
                      resource=resource,
                      pk=pk,
                      entity_refs=entity_refs,
                      count=count,
                      check_ordering=check_ordering)


def assert_201(response, vnd=True, entity_refs=None, included=None, is_list=False,
               resource=None, pk=None, attributes=None, relationships=None):
    assert response is not None
    assert response.status_code == status.HTTP_201_CREATED, f'status_code {response.status_code} != 201'
    response_has_data(response,
                      attributes=attributes,
                      relationships=relationships,
                      included=included,
                      is_list=is_list,
                      vnd=vnd,
                      resource=resource,
                      pk=pk,
                      entity_refs=entity_refs)


def assert_202(response, vnd=True, entity_refs=None, included=None, is_list=False,
               resource=None, pk=None, attributes=None, relationships=None):
    assert response is not None
    assert response.status_code == status.HTTP_202_ACCEPTED, f'status_code {response.status_code} != 202'
    response_has_data(response,
                      attributes=attributes,
                      relationships=relationships,
                      included=included,
                      is_list=is_list,
                      vnd=vnd,
                      resource=resource,
                      pk=pk,
                      entity_refs=entity_refs)


def assert_204(response):
    assert response is not None
    assert response.status_code == status.HTTP_204_NO_CONTENT, f'status_code {response.status_code} != 204'


def assert_400(response, error=None, pointer=None, vnd=True):
    assert response is not None
    assert response.status_code == status.HTTP_400_BAD_REQUEST, f'status_code {response.status_code} != 400'
    response_has_error(response, error, pointer, vnd)


def assert_401(response, error='Authentication credentials were not provided', vnd=True):
    assert response is not None
    assert response.status_code == status.HTTP_401_UNAUTHORIZED, f'status_code {response.status_code} != 401'
    response_has_error(response, error, vnd=vnd)


def assert_403(response, error='You do not have permission to perform this action', vnd=True):
    assert response is not None
    assert response.status_code == status.HTTP_403_FORBIDDEN, f'status_code {response.status_code} != 403'
    response_has_error(response, error, vnd=vnd)


def assert_404(response, error='Not found', pointer=None, vnd=True):
    assert response is not None
    assert response.status_code == status.HTTP_404_NOT_FOUND, f'status_code {response.status_code} != 404'
    response_has_error(response, error, pointer, vnd)


def assert_405(response, error='Method not allowed', pointer=None, vnd=True):
    assert response is not None
    assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED, f'status_code {response.status_code} != 405'
    response_has_error(response, error, pointer, vnd)


def assert_500(response, error='A server error occurred.', pointer=None):
    assert response is not None
    assert response.status_code == status.HTTP_500_INTERNAL_SERVER_ERROR, f'status_code {response.status_code} != 500'
    response_has_error(response, error, pointer)


def assert_503(response, error='Service temporarily unavailable, try again later', pointer=None):
    assert response is not None
    assert response.status_code == status.HTTP_503_SERVICE_UNAVAILABLE, f'status_code {response.status_code} != 503'
    response_has_error(response, error, pointer)


class AssertionHelper:
    EntityRef = EntityReferenceClass

    HTTP_200 = assert_200
    HTTP_201 = assert_201
    HTTP_202 = assert_202
    HTTP_204 = assert_204

    HTTP_400 = assert_400
    HTTP_401 = assert_401
    HTTP_403 = assert_403
    HTTP_404 = assert_404
    HTTP_405 = assert_405

    HTTP_500 = assert_500
    HTTP_503 = assert_503


@pytest.fixture(scope='session')
def json_asserter():
    return AssertionHelper


class JsonAsserterMixin:
    @pytest.fixture(autouse=True)
    def set_json_asserter(self, json_asserter):
        self.json_asserter = json_asserter



"""
Copyright 2019 ShipChain, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from rest_framework import permissions


class HasViewSetActionPermissions(permissions.BasePermission):
    def has_permission(self, request, view):
        """
        Look for list of required FeaturePermissions in view.action_user_permissions, which
        should be a dict mapping action name (key) to FeaturePermission (value),
        i.e.:

        class MyViewSet(HasViewSetActionPermissions, ViewSet):
            permission_classes = (permissions.IsAuthenticated, HasViewSetActionPermissions)
            action_user_permissions = {
                'create': 'my_feature.create_object',
                'retrieve': 'my_feature.view_object',
                'my_action': 'my_feature.special_permission',
            }

            @action
            def my_action:
                ...

        If there's no entry for that action then return True so other Permission checks can run.

        This expects User.has_perm() to be provided by the authentication backend/class

        Built-in actions:
            create
            retrieve
            update         PUT
            partial_update PATCH
            destroy
            list
        """
        if request.user and request.user.is_authenticated:

            if view.action_user_permissions is None:
                return True

            action = view.action

            # PUT and PATCH should use the same permissions if not separately defined
            if action == 'partial_update' and 'partial_update' not in view.action_user_permissions:
                action = 'update'

            try:
                perms = view.action_user_permissions[action]
                if isinstance(perms, str):
                    perms = [perms]
            except (KeyError, AttributeError):
                return True

            return request.user.has_perms(perms)
        return False



"""
Copyright 2019 ShipChain, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import decimal
import json
import re
from uuid import UUID

from dateutil.parser import parse
from django.db import models
from django.conf import settings
from django.core.mail import EmailMessage
from django.template.loader import render_to_string
from enumfields.drf import EnumField
from rest_framework.exceptions import ValidationError


DN_REGEX = re.compile(r'(?:/?)(.+?)(?:=)([^/]+)')


def assertDeepAlmostEqual(test_case, expected, actual, *args, **kwargs):  # nopep8 pylint: disable=invalid-name
    """
    Assert that two complex structures have almost equal contents.

    Compares lists, dicts and tuples recursively. Checks numeric values
    using test_case's :py:meth:`unittest.TestCase.assertAlmostEqual` and
    checks all other values with :py:meth:`unittest.TestCase.assertEqual`.
    Accepts additional positional and keyword arguments and pass those
    intact to assertAlmostEqual() (that's how you specify comparison
    precision).

    :param test_case: TestCase object on which we can call all of the basic
    'assert' methods.
    :type test_case: :py:class:`unittest.TestCase` object
    """
    is_root = '__trace' not in kwargs
    trace = kwargs.pop('__trace', 'ROOT')
    try:
        if isinstance(expected, (int, float, int, complex)):
            test_case.assertAlmostEqual(expected, actual, *args, **kwargs)
        elif isinstance(expected, dict):
            test_case.assertEqual(set(expected), set(actual))
            for key in expected:
                assertDeepAlmostEqual(test_case, expected[key], actual[key],
                                      __trace=repr(key), *args, **kwargs)
        else:
            test_case.assertEqual(expected, actual)
    except AssertionError as exc:
        exc.__dict__.setdefault('traces', []).append(trace)
        if is_root:
            trace = ' -> '.join(reversed(exc.traces))
            exc = AssertionError("%s\nTRACE: %s" % (str(exc), trace))
        raise exc


def build_auth_headers_from_request(request):
    if not request.auth or not isinstance(request.auth, bytes):
        raise Exception("No auth in request")

    token = request.auth.decode('utf-8')
    return {'Authorization': f"JWT {token}"}


def get_client_ip(request):
    """
    Returns ip address from which the http request originated
    """
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        addr = x_forwarded_for.split(',')[0]
    else:
        addr = request.META.get('REMOTE_ADDR')
    return addr


def get_domain_from_email(email):
    if '@' in email:
        domain = email.split("@")[1]
        return domain
    raise ValidationError("Domain can't be determined from username.")


def is_date(string, fuzzy=False):
    """
    Return whether the string can be interpreted as a date.

    :param string: str, string to check for date
    :param fuzzy: bool, ignore unknown tokens in string if True
    """
    try:
        parse(string, fuzzy=fuzzy)
        return True

    except ValueError:
        return False


def tznow():
    """
    Return an aware or naive datetime.datetime, depending on settings.USE_TZ.
    """
    from django.utils.timezone import now
    return now()


def random_id():
    """
    Cast the UUID to a string
    """
    from uuid import uuid4
    return str(uuid4())


def validate_uuid4(uuid_string):
    """
    Validate that a UUID string is in fact a valid uuid4.
    Happily, the uuid module does the actual checking for us.
    It is vital that the 'version' kwarg be passed to the UUID() call
    otherwise any 32-character hex string is considered valid.
    """

    try:
        val = UUID(uuid_string, version=4)
    except ValueError:
        # If it's a value error, then the string
        # is not a valid hex code for a UUID.
        return False

    # Check to make sure that the uuid has dashes.
    return str(val) == uuid_string


def remove_dict_key_recursively(dict_obj, list_key_to_remove):
    """
    :param dict_obj: Dictionary object from which we want to remove a particular key
    :param list_key_to_remove: A list of key to remove recursively
    :return: The input dict_obj without the key_to_remove key if found
    """
    keys_to_remove = [key.lower() for key in list_key_to_remove]
    to_return = {}
    for key, value in dict_obj.items():
        if key.lower() not in keys_to_remove:
            if isinstance(value, dict):
                to_return[key] = remove_dict_key_recursively(value, keys_to_remove)
            else:
                to_return[key] = value

    return to_return


def send_templated_email(template, subject, context, recipients, sender=None):
    request = context.get('request', None)
    send_by = sender if sender else settings.DEFAULT_FROM_EMAIL
    email_body = render_to_string(template, context=context, request=request)
    email = EmailMessage(subject, email_body, send_by, recipients)
    email.content_subtype = 'html'
    email.send()


def snake_to_sentence(word):
    return ' '.join(x.capitalize() or '_' for x in word.split('_'))


def parse_dn(ssl_dn):
    return dict(DN_REGEX.findall(ssl_dn))


class AliasField(models.Field):
    def contribute_to_class(self, cls, name, private_only=False):
        """
            virtual_only is deprecated in favor of private_only
        """
        super(AliasField, self).contribute_to_class(cls, name, private_only=True)
        setattr(cls, name, self)

    def __get__(self, instance, instance_type=None):
        return getattr(instance, self.db_column)


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):  # pylint: disable=method-hidden
        if isinstance(o, decimal.Decimal):
            return float(o)
        return super(DecimalEncoder, self).default(o)


class EnumIntegerFieldLabel(EnumField):
    def to_representation(self, instance):
        return str(instance)


class UpperEnumField(EnumField):
    def to_representation(self, instance):
        return super(UpperEnumField, self).to_representation(instance).upper()



from collections import OrderedDict
from rest_framework.fields import SkipField


class NullableFieldsMixin:
    def to_representation(self, instance):
        # Remove null fields from serialized object
        ret = OrderedDict()
        fields = [field for field in self.fields.values() if not field.write_only]

        for field in fields:
            try:
                attribute = field.get_attribute(instance)
            except SkipField:
                continue

            if attribute is not None:
                representation = field.to_representation(attribute)
                if representation is None:
                    # Do not serialize empty objects
                    continue
                if isinstance(representation, list) and not representation:
                    # Do not serialize empty lists
                    continue
                ret[field.field_name] = representation

        return ret



"""
Copyright 2019 ShipChain, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from rest_framework import status, mixins
from rest_framework.response import Response


class SerializationType:
    REQUEST = 0
    RESPONSE = 1


class MultiSerializerViewSetMixin:
    def get_serializer_class(self):
        """
        Look for serializer class in self.action_serializer_classes, which
        should be a dict mapping action name (key) to serializer class (value),
        i.e.:

        class MyViewSet(MultiSerializerViewSetMixin, ViewSet):
            serializer_class = MyDefaultSerializer
            action_serializer_classes = {
               'list': MyListSerializer,
               'my_action': MyActionSerializer,
            }

            @action
            def my_action:
                ...

        If there's no entry for that action then just fallback to the regular
        get_serializer_class lookup: self.serializer_class, DefaultSerializer.

        Built-in actions:
            create
            retrieve
            update         PUT
            partial_update PATCH
            destroy
            list
        """
        action = self.action

        try:
            # PUT and PATCH should use the same serializers if not separately defined
            if action == 'partial_update' and 'partial_update' not in self.action_serializer_classes:
                action = 'update'

            if 'format' in self.kwargs:
                try:
                    return self.action_serializer_classes[f'{action}.{self.kwargs["format"]}']
                except (KeyError, AttributeError):
                    pass

            return self.action_serializer_classes[action]

        except (KeyError, AttributeError):
            return super(MultiSerializerViewSetMixin, self).get_serializer_class()


class MultiPermissionViewSetMixin:
    def get_permissions(self):
        """
        Look for permission classes in self.action_permission_classes, which
        should be a dict mapping action name (key) to list of permission classes (value),
        i.e.:

        class MyViewSet(MultiPermissionViewSetMixin, ViewSet):
            permission_classes = (HasViewSetActionPermissions,)
            action_permission_classes = {
               'list': (permissions.AllowAny,),
               'my_action': (HasViewSetActionPermissions, isOwner,),
            }

            @action
            def my_action:
                ...

        If there's no entry for that action then just fallback to the regular permission_classes

        Built-in actions:
            create
            retrieve
            update         PUT
            partial_update PATCH
            destroy
            list
        """
        permission_classes = self.permission_classes

        try:
            action = self.action

            # PUT and PATCH should use the same permission classes if not separately defined
            if action == 'partial_update' and 'partial_update' not in self.action_permission_classes:
                action = 'update'

            permission_classes = self.action_permission_classes[action]

        except (KeyError, AttributeError):
            pass

        return [permission() for permission in permission_classes]


class ConfigurableCreateModelMixin(mixins.CreateModelMixin):
    def create(self, request, *args, **kwargs):
        """
        Provide SerializationType to get_serializer, re-serialize response if necessary, change status code if wanted
        """
        config = self.get_configuration()

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=config.raise_validation)
        instance = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)

        if config.re_serialize_response:
            serializer = self.get_serializer(instance=instance, serialization_type=SerializationType.RESPONSE)

        return Response(serializer.data,
                        status=config.success_status or status.HTTP_201_CREATED,
                        headers=headers)

    def perform_create(self, serializer):
        """Returned the saved instance for later processing"""
        return serializer.save()


class ConfigurableRetrieveModelMixin(mixins.RetrieveModelMixin):
    """
    Retrieve a model instance.  Provide SerializationType to get_serializer
    """
    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance=instance, serialization_type=SerializationType.RESPONSE)
        return Response(serializer.data)


class ConfigurableUpdateModelMixin(mixins.UpdateModelMixin):
    def update(self, request, *args, **kwargs):
        """
        Provide SerializationType to get_serializer, re-serialize response if necessary, change status code if wanted
        """
        config = self.get_configuration()

        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        serializer = self.get_serializer(instance=instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=config.raise_validation)
        updated_instance = self.perform_update(serializer)

        if getattr(instance, '_prefetched_objects_cache', None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}  # pylint: disable=protected-access

        if config.re_serialize_response:
            serializer = self.get_serializer(instance=updated_instance, serialization_type=SerializationType.RESPONSE)

        return Response(serializer.data, status=config.success_status or status.HTTP_200_OK)

    def perform_update(self, serializer):
        """Returned the updated instance for later processing"""
        return serializer.save()


class ConfigurableDestroyModelMixin(mixins.DestroyModelMixin):
    """No specific Configurable changes; creating mixin for consistency"""


class ConfigurableListModelMixin(mixins.ListModelMixin):
    """
    List a queryset.  Provide SerializationType to get_serializer
    """
    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True, serialization_type=SerializationType.RESPONSE)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True, serialization_type=SerializationType.RESPONSE)
        return Response(serializer.data)
